from .lib import dll, EcosLib
from .EcosSimulation import EcosSimulation

from ctypes import c_void_p, c_double


class EcosSimulationRunner:
    """
    This class is used to run a simulation.
    """

    def __init__(self, sim: EcosSimulation):

        _simulation_runner_create = dll.ecos_simulation_runner_create
        _simulation_runner_create.restype = c_void_p
        _simulation_runner_create.argtypes = [c_void_p]

        self._handle = _simulation_runner_create(sim.sim)
        if self._handle is None:
            raise Exception(EcosLib.get_last_error())

    def start(self):
        """
        Starts simulation, which runs in a separate native thread.
        """
        _simulation_runner_start = dll.ecos_simulation_runner_start
        _simulation_runner_start.argtypes = [c_void_p]
        _simulation_runner_start(self._handle)

    def stop(self):
        """
        Stops the simulation. This will stop the simulation thread.
        """
        _simulation_runner_stop = dll.ecos_simulation_runner_stop
        _simulation_runner_stop.argtypes = [c_void_p]
        _simulation_runner_stop(self._handle)

    def set_real_time_factor(self, factor: float):
        """
        Sets the real-time factor for the simulation. This factor determines how fast the simulation should run compared to real time.
        Args:
            factor (float): The real-time factor. A value of 1.0 means real-time, 2.0 means twice as fast, etc.
        """
        _simulation_runner_set_real_time_factor = dll.ecos_simulation_runner_set_real_time_factor
        _simulation_runner_set_real_time_factor.argtypes = [c_void_p, c_double]
        _simulation_runner_set_real_time_factor(self._handle, factor)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.free()

    def free(self):
        """
        Frees the resources associated with the simulation structure.
        """
        if not self._handle is None:
            destroy_simulation_runner = dll.ecos_simulation_runner_destroy
            destroy_simulation_runner.argtypes = [c_void_p]
            destroy_simulation_runner(self._handle)
            self._handle = None

    def __del__(self):
        self.free()



from .lib import dll, EcosLib
from .EcosSimulationStructure import EcosSimulationStructure
from ctypes import c_bool, c_int, c_void_p, c_double, c_char_p, c_size_t, POINTER, Structure, CFUNCTYPE, byref, \
    create_string_buffer


class SimulationInfo(Structure):
    _fields_ = [("time", c_double), ("iterations", c_size_t)]


class SimulationListener:

    def pre(self, info: SimulationInfo):
        pass

    def post(self, info: SimulationInfo):
        pass


class ListenerConfig(Structure):
    _fields_ = [
        ("preStepCallback", CFUNCTYPE(None, SimulationInfo)),
        ("postStepCallback", CFUNCTYPE(None, SimulationInfo))
    ]


class EcosSimulation:
    """
    This class represents a co-simulation.
    Supports context management for automatic resource management.
    """

    def __init__(self, step_size: float, ssp_path: str = None, structure: EcosSimulationStructure = None):
        """
        Initialize a new EcosSimulation instance.

        Args:
           step_size (float): The fixed simulation step size.
           ssp_path (str, optional): Path to an SSP file to create the simulation from. Mutually exclusive with `structure`.
           structure (EcosSimulationStructure, optional): Simulation structure to create the simulation from. Mutually exclusive with `ssp_path`.

        Raises:
           ValueError: If both or neither of `ssp_path` and `structure` are provided.
           Exception: If simulation creation fails in the underlying library.
        """
        if ssp_path is not None and structure is not None:
            raise ValueError("Only one of 'ssp_path' or 'structure' should be provided.")

        self.listener_configs = {}

        self._init_simulation = dll.ecos_simulation_init
        self._init_simulation.argtypes = [c_void_p, c_double, c_char_p]
        self._init_simulation.restype = c_bool

        self._step_simulation = dll.ecos_simulation_step
        self._step_simulation.argtypes = [c_void_p, c_size_t]
        self._step_simulation.restype = c_double

        self._step_simulation_for = dll.ecos_simulation_step_for
        self._step_simulation_for.argtypes = [c_void_p, c_double]

        self._step_simulation_until = dll.ecos_simulation_step_until
        self._step_simulation_until.argtypes = [c_void_p, c_double]

        self._get_time = dll.ecos_simulation_get_time
        self._get_time.argtypes = [c_void_p]
        self._get_time.restype = c_double

        self._get_integer = dll.ecos_simulation_get_integer
        self._get_integer.argtypes = [c_void_p, c_char_p, POINTER(c_int)]
        self._get_integer.restype = c_bool

        self._get_real = dll.ecos_simulation_get_real
        self._get_real.argtypes = [c_void_p, c_char_p, POINTER(c_double)]
        self._get_real.restype = c_bool

        self._get_bool = dll.ecos_simulation_get_bool
        self._get_bool.argtypes = [c_void_p, c_char_p, POINTER(c_bool)]
        self._get_bool.restype = c_bool

        self._get_string = dll.ecos_simulation_get_string
        self._get_string.argtypes = [c_void_p, c_char_p, c_char_p]
        self._get_string.restype = c_bool

        self._set_integer = dll.ecos_simulation_set_integer
        self._set_integer.argtypes = [c_void_p, c_char_p, c_int]
        self._set_integer.restype = c_bool

        self._set_real = dll.ecos_simulation_set_real
        self._set_real.argtypes = [c_void_p, c_char_p, c_double]
        self._set_real.restype = c_bool

        self._set_bool = dll.ecos_simulation_set_bool
        self._set_bool.argtypes = [c_void_p, c_char_p, c_bool]
        self._set_bool.restype = c_bool

        self._set_string = dll.ecos_simulation_set_string
        self._set_string.argtypes = [c_void_p, c_char_p, c_char_p]
        self._set_string.restype = c_bool

        self._create_listener = dll.ecos_simulation_listener_create
        self._create_listener.argtypes = [ListenerConfig]
        self._create_listener.restype = c_void_p

        self._create_csv_writer = dll.ecos_csv_writer_create
        self._create_csv_writer.argtypes = [c_char_p, c_char_p]
        self._create_csv_writer.restype = c_void_p

        self._add_listener = dll.ecos_simulation_add_listener
        self._add_listener.argtypes = [c_void_p, c_char_p, c_void_p]

        self._remove_listener = dll.ecos_simulation_remove_listener
        self._remove_listener.argtypes = [c_void_p, c_char_p]

        if ssp_path is not None:

            _create_simulation_from_ssp = dll.ecos_simulation_create_from_ssp
            _create_simulation_from_ssp.argtypes = [c_char_p, c_double]
            _create_simulation_from_ssp.restype = c_void_p

            self.sim = _create_simulation_from_ssp(ssp_path.encode(), step_size)
            if self.sim is None:
                raise Exception(EcosLib.get_last_error())

        elif structure is not None:
            _create_simulation_from_structure = dll.ecos_simulation_create_from_structure
            _create_simulation_from_structure.argtypes = [c_void_p, c_double]
            _create_simulation_from_structure.restype = c_void_p

            self.sim = _create_simulation_from_structure(structure.handle, step_size)
            if self.sim is None:
                raise Exception(EcosLib.get_last_error())

        else:
            raise ValueError("Either 'ssp_path' or 'structure' must be provided.")

    def add_listener(self, name: str, listener: SimulationListener):
        """
        Add a listener to the simulation.

        Args:
            name (str): name of the listener (used to identify the listener)
            listener (SimulationListener): a listener object that implements the SimulationListener interface
        """
        if not isinstance(listener, SimulationListener):
            raise Exception("listener must be of type SimulationListener")

        listener.name = name

        config = ListenerConfig()
        self.listener_configs[name] = config

        @CFUNCTYPE(None, SimulationInfo)
        def pre(info: SimulationInfo):
            listener.pre(info)

        config.preStepCallback = pre

        @CFUNCTYPE(None, SimulationInfo)
        def post(info: SimulationInfo):
            listener.post(info)

        config.postStepCallback = post

        cpp_listener = self._create_listener(config)
        self._add_listener(self.sim, name.encode(), cpp_listener)

    def remove_listener(self, name: str):
        """
        Remove a listener from the simulation.
        Args:
            name (str): Name of the listener to remove.
        """
        self._remove_listener(self.sim, name.encode())

    def add_scenario(self, scenario_config: str = None, real_actions: list[tuple] = None, integer_actions: list[tuple]  = None,
                     boolean_actions: list[tuple]  = None,
                     string_actions: list[tuple] =None):
        """
        Add a scenario to the simulation.
        Args:
            scenario_config (str, optional): Optional scenario configuration XML file.
            real_actions (list of tuples): List of real actions as (time_point, identifier, value).
            integer_actions (list of tuples): List of integer actions as (time_point, identifier, value).
            boolean_actions (list of tuples): List of boolean actions as (time_point, identifier, value).
            string_actions (list of tuples): List of string actions as (time_point, identifier, value).
        """

        listener = None
        if scenario_config is not None:
            load_scenario = dll.ecos_scenario_load
            load_scenario.argtypes = [c_char_p]
            load_scenario.restype = c_void_p
            listener = load_scenario(str.encode("utf-8"))

        else:
            create_scenario = dll.ecos_scenario_create
            create_scenario.restype = c_void_p
            listener = create_scenario()

        if listener is None:
            raise ValueError("Unable to create/load scenario: " + EcosLib.get_last_error())

        if real_actions is not None:
            for action in real_actions:
                scenario_add_real_action = dll.ecos_scenario_add_real_action
                scenario_add_real_action.argtypes = [c_void_p, c_double, c_char_p, c_double]
                scenario_add_real_action.restype = c_bool

                time_point, identifier, value = action
                scenario_add_real_action(listener, time_point, identifier.encode("utf-8"), value)

        if integer_actions is not None:
            for action in integer_actions:
                scenario_add_integer_action = dll.ecos_scenario_add_real_action
                scenario_add_integer_action.argtypes = [c_void_p, c_double, c_char_p, c_int]
                scenario_add_integer_action.restype = c_bool

                time_point, identifier, value = action
                scenario_add_integer_action(listener, time_point, identifier.encode("utf-8"), value)

        if boolean_actions is not None:
            for action in boolean_actions:
                scenario_add_boolean_action = dll.ecos_scenario_add_real_action
                scenario_add_boolean_action.argtypes = [c_void_p, c_double, c_char_p, c_bool]
                scenario_add_boolean_action.restype = c_bool

                time_point, identifier, value = action
                scenario_add_boolean_action(listener, time_point, identifier.encode("utf-8"), value)

        if string_actions is not None:
            for action in string_actions:
                scenario_add_string_action = dll.ecos_scenario_add_real_action
                scenario_add_string_action.argtypes = [c_void_p, c_double, c_char_p, c_char_p]
                scenario_add_string_action.restype = c_bool

                time_point, identifier, value = action
                scenario_add_string_action(listener, time_point, identifier.encode("utf-8"), value.encode("utf-8"))

        self._add_listener(self.sim, b'scenario', listener)

    def add_csv_writer(self, result_file: str, csv_config: str = None, identifiers: list[str] = None,
                       decimation_factor: int = None):
        """
        Add a CSV writer to the simulation.
        Args:
            result_file (str): Path to the CSV result file.
            csv_config (str, optional): Optional CSV configuration string.
            identifiers (list[str], optional): List of variable identifiers to record.
            decimation_factor (int, optional): Factor to reduce the number of recorded samples.
        """

        listener = self._create_csv_writer(result_file.encode(),
                                           None if csv_config is None else csv_config.encode())
        if listener is None:
            raise Exception(EcosLib.get_last_error())

        if identifiers is not None:
            register_variable = dll.ecos_csv_writer_register_variable
            register_variable.argtypes = [c_void_p, c_char_p]
            register_variable.restype = c_bool

            for identifier in identifiers:
                if not register_variable(listener, identifier.encode()):
                    raise Exception(EcosLib.get_last_error())

        if decimation_factor is not None:
            set_decimation = dll.ecos_csv_writer_set_decimation_factor
            set_decimation.argtypes = [c_void_p, c_int]
            set_decimation.restype = c_bool

            if not set_decimation(listener, decimation_factor):
                raise Exception(EcosLib.get_last_error())

        self._add_listener(self.sim, b'csv_writer', listener)

    def get_integer(self, identifier: str):
        val = c_int()
        if not self._get_integer(self.sim, identifier.encode(), byref(val)):
            raise Exception(EcosLib.get_last_error())
        return val.value

    def get_real(self, identifier: str):
        val = c_double()
        if not self._get_real(self.sim, identifier.encode(), byref(val)):
            raise Exception(EcosLib.get_last_error())
        return val.value

    def get_bool(self, identifier: str):
        val = c_bool()
        if not self._get_bool(self.sim, identifier.encode(), byref(val)):
            raise Exception(EcosLib.get_last_error())
        return val.value

    def get_string(self, identifier: str, buffer_size: int = 1024):
        buffer = create_string_buffer(buffer_size)
        if not self._get_string(self.sim, identifier.encode(), buffer):
            raise Exception(EcosLib.get_last_error())
        return buffer.value.decode()

    def set_integer(self, identifier: str, value: int):
        if not self._set_integer(self.sim, identifier.encode(), value):
            raise Exception(EcosLib.get_last_error())

    def set_real(self, identifier: str, value: float):
        if not self._set_real(self.sim, identifier.encode(), value):
            raise Exception(EcosLib.get_last_error())

    def set_bool(self, identifier: str, value: bool):
        if not self._set_bool(self.sim, identifier.encode(), value):
            raise Exception(EcosLib.get_last_error())

    def set_string(self, identifier: str, value: str):
        if not self._set_string(self.sim, identifier.encode(), value.encode()):
            raise Exception(EcosLib.get_last_error())

    def init(self, start_time: int = 0, parameter_set: str = None):
        """
        Initialize the simulation with a start time and an optional parameter set.
        Args:
            start_time (int): The time to start the simulation from in seconds. Defaults to 0.
            parameter_set (str, optional): Optional parameter set to initialize the simulation with.
        """
        return self._init_simulation(self.sim, start_time, None if parameter_set is None else parameter_set.encode())

    def step(self, num_steps: int = 1) -> float:
        """
        Step the simulation for a given number of steps.
        Args:
            num_steps (int): Number of steps to take in the simulation. Defaults to 1.
        Returns:
            float: The simulation time after stepping.
        """
        return self._step_simulation(self.sim, num_steps)

    def step_for(self, duration: float):
        """
        Step the simulation for a specified duration.
        Args:
            duration (float): Duration to step the simulation for in seconds.
        """
        self._step_simulation_for(self.sim, duration)

    def step_until(self, time_point: float):
        """
        Step the simulation until a specified time point.
        Args:
            time_point (float): The time point to step the simulation until in seconds.
        """
        self._step_simulation_until(self.sim, time_point)

    def get_time(self) -> float:
        """
        Get the current simulation time.
        Returns:
            float: The current simulation time in seconds.
        """
        return self._get_time(self.sim)

    def reset(self):
        """
        Reset the simulation to its initial state.
        Raises:
            Exception: If the reset operation fails in the underlying library.
        """
        reset_simulation = dll.ecos_simulation_reset
        reset_simulation.argtypes = [c_void_p]
        reset_simulation.restype = c_bool

        if not reset_simulation(self.sim):
            raise Exception(EcosLib.get_last_error())

    def terminate(self):
        """
        Terminate the simulation gracefully.
        Raises:
            Exception: If the termination operation fails in the underlying library.
        """
        terminate_simulation = dll.ecos_simulation_terminate
        terminate_simulation.argtypes = [c_void_p]
        terminate_simulation.restype = c_bool

        if not terminate_simulation(self.sim):
            raise Exception(EcosLib.get_last_error())

    def free(self):
        """
        Frees the resources associated with the simulation structure.
        """
        if not self.sim is None:
            destroy_simulation = dll.ecos_simulation_destroy
            destroy_simulation.argtypes = [c_void_p]

            destroy_simulation(self.sim)
            self.sim = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.free()

    def __del__(self):
        self.free()



import os
from .EcosParameterSet import EcosParameterSet
from .lib import dll, EcosLib
from ctypes import c_void_p, c_bool, c_char_p, CFUNCTYPE, c_double


class EcosSimulationStructure:
    """
    Represents the structure of a simulation.

    Can be passed to EcosSimulation to create a simulation instance.

    Supports context management for automatic resource management.
    """

    def __init__(self):

        self.modifiers = []  # to keep modifier function alive

        create_simulation_structure = dll.ecos_simulation_structure_create
        create_simulation_structure.restype = c_void_p

        self._add_model = dll.ecos_simulation_structure_add_model
        self._add_model.argtypes = [c_void_p, c_char_p, c_char_p, c_double]
        self._add_model.restype = c_bool

        self._add_parameter_set = dll.ecos_simulation_structure_add_parameter_set
        self._add_parameter_set.argtypes = [c_void_p, c_char_p, c_void_p]
        self._add_parameter_set.restype = c_bool

        self._make_int_connection = dll.ecos_simulation_structure_make_int_connection
        self._make_int_connection.argtypes = [c_void_p, c_char_p, c_char_p]

        self._make_real_connection_mod = dll.ecos_simulation_structure_make_real_connection_mod
        self._make_real_connection_mod.argtypes = [c_void_p, c_char_p, c_char_p, CFUNCTYPE(c_double, c_double)]

        self._make_real_connection = dll.ecos_simulation_structure_make_real_connection
        self._make_real_connection.argtypes = [c_void_p, c_char_p, c_char_p]

        self._make_string_connection = dll.ecos_simulation_structure_make_string_connection
        self._make_string_connection.argtypes = [c_void_p, c_char_p, c_char_p]

        self._make_bool_connection = dll.ecos_simulation_structure_make_bool_connection
        self._make_bool_connection.argtypes = [c_void_p, c_char_p, c_char_p]

        self._handle = create_simulation_structure()
        if self._handle is None:
            raise Exception(EcosLib.get_last_error())

    def add_model(self, instance_name: str, uri: str | os.PathLike, step_size_hint: float = None):
        """
        Adds a model to the simulation structure.
        Args:
            instance_name (str): The name of the model instance.
            uri (str | os.PathLike): The URI or path to the model file.
            step_size_hint (float, optional): A hint for the step size. Defaults to None.
        Raises:
            Exception: If adding the model fails.
        """
        if isinstance(uri, os.PathLike):
            uri = os.fspath(uri)

        if not self._add_model(self.handle, instance_name.encode(), uri.encode(),
                               -1 if step_size_hint is None else step_size_hint):
            raise Exception(EcosLib.get_last_error())

    def add_parameter_set(self, name: str, parameters: EcosParameterSet | dict[str, any]):
        """
        Adds a parameter set to the simulation structure.
        Args:
            name (str): The name of the parameter set.
            parameters (EcosParameterSet | dict[str, any]): The parameters to add, either as an EcosParameterSet or a dictionary.
        Raises:
            Exception: If the parameters are of an illegal type or if adding the parameter set fails.
        """
        if isinstance(parameters, EcosParameterSet):
            return self._add_parameter_set(self.handle, name.encode(), parameters.handle)
        elif isinstance(parameters, dict):
            with (EcosParameterSet()) as params:
                for key, value in parameters.items():
                    if isinstance(value, bool):
                        params.add_bool(key, value)
                    elif isinstance(value, float):
                        params.add_real(key, value)
                    elif isinstance(value, int):
                        params.add_int(key, value)
                    elif isinstance(value, str):
                        params.add_string(key, value)
                    else:
                        raise Exception("Illegal value type. Must be int, float, bool or str")
                return self.add_parameter_set(name, params)

        else:
            raise Exception("Illegal parameter type. Must be EcosParameterSet or dict")

    def make_real_connection(self, source: str, sink: str, modifier=None):
        """
        Creates a connection between two real variables in the simulation structure.
        Args:
            source (str): The source variable identifier.
            sink (str): The sink variable identifier.
            modifier (callable, optional): A function to modify the value before passing it to the sink. Defaults to None.
        """
        if modifier is not None:
            @CFUNCTYPE(c_double, c_double)
            def _modifier(value: float) -> float:
                return modifier(value)

            self.modifiers.append(_modifier)
            self._make_real_connection_mod(self.handle, source.encode(), sink.encode(), _modifier)

        else:
            self._make_real_connection(self.handle, source.encode(), sink.encode())

    def make_int_connection(self, source: str, sink: str):
        """
        Creates a connection between two integer variables in the simulation structure.
        Args:
            source (str): The source variable identifier.
            sink (str): The sink variable identifier.
        """
        self._make_int_connection(self.handle, source.encode(), sink.encode())

    def make_string_connection(self, source: str, sink: str):
        """
       Creates a connection between two string variables in the simulation structure.
       Args:
           source (str): The source variable identifier.
           sink (str): The sink variable identifier.
       """
        self._make_string_connection(self.handle, source.encode(), sink.encode())

    def make_bool_connection(self, source: str, sink: str):
        """
       Creates a connection between two boolean variables in the simulation structure.
       Args:
           source (str): The source variable identifier.
           sink (str): The sink variable identifier.
       """
        self._make_bool_connection(self.handle, source.encode(), sink.encode())

    @property
    def handle(self):
        return self._handle

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.free()

    def free(self):
        """
        Frees the resources associated with the simulation structure.
        """
        if not self._handle is None:
            simulation_structure_destroy = dll.ecos_simulation_structure_destroy
            simulation_structure_destroy.argtypes = [c_void_p]

            simulation_structure_destroy(self._handle)
            self._handle = None
            self.modifiers.clear()

    def __del__(self):
        self.free()



from ctypes import c_void_p, c_char_p, c_int, c_bool, c_double

from .lib import dll, EcosLib


class EcosParameterSet:
    """
    This class represents a set of parameter values that can be passed to a simulation.
    Supports context management for automatic resource management.
    """

    def __init__(self):
        self._parameter_set_add_int = dll.ecos_parameter_set_add_int
        self._parameter_set_add_int.argtypes = [c_void_p, c_char_p, c_int]

        self._parameter_set_add_real = dll.ecos_parameter_set_add_real
        self._parameter_set_add_real.argtypes = [c_void_p, c_char_p, c_double]

        self._parameter_set_add_string = dll.ecos_parameter_set_add_string
        self._parameter_set_add_string.argtypes = [c_void_p, c_char_p, c_char_p]

        self._parameter_set_add_bool = dll.ecos_parameter_set_add_bool
        self._parameter_set_add_bool.argtypes = [c_void_p, c_char_p, c_bool]

        create_parameter_set = dll.ecos_parameter_set_create
        create_parameter_set.restype = c_void_p
        self._handle = create_parameter_set()
        if self._handle is None:
            raise Exception(EcosLib.get_last_error())

    def add_int(self, name: str, value: int):
        self._parameter_set_add_int(self.handle, name.encode(), value)

    def add_real(self, name: str, value: float):
        self._parameter_set_add_real(self.handle, name.encode(), value)

    def add_bool(self, name: str, value: bool):
        self._parameter_set_add_bool(self.handle, name.encode(), value)

    def add_string(self, name: str, value: str):
        self._parameter_set_add_string(self.handle, name.encode(), value.encode())

    @property
    def handle(self):
        return self._handle

    def free(self):
        """
        Frees the resources associated with the simulation structure.
        """
        if not self._handle is None:
            destroy_parameter_set = dll.ecos_parameter_set_destroy
            destroy_parameter_set.argtypes = [c_void_p]

            destroy_parameter_set(self._handle)
            self._handle = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.free()

    def __del__(self):
        self.free()



import os
from pathlib import Path
from ctypes import CDLL, Structure, c_char_p, c_int


def load_library():
    def suffix() -> str:
        return ".dll" if os.name == "nt" else ".so"

    bin_folder = str((Path(__file__).parent / 'binaries').resolve())
    lib_name = f"libecosc{suffix()}"

    current_path = os.environ.get("PATH", "")

    # Add the executable directory to the PATH if not already present
    if bin_folder not in current_path:
        os.environ["PATH"] = bin_folder + os.pathsep + current_path

    if os.name == "nt":
        with os.add_dll_directory(bin_folder):
            return CDLL(lib_name)
    else:
        return CDLL(f"{bin_folder}/{lib_name}")


dll = load_library()


class Version(Structure):
    _fields_ = [("major", c_int), ("minor", c_int), ("patch", c_int)]

    def __repr__(self):
        return f"v{self.major}.{self.minor}.{self.patch}"


class EcosLib:

    @staticmethod
    def version():
        version_fun = dll.ecos_library_version
        version_fun.restype = Version
        return dll.ecos_library_version()

    @staticmethod
    def get_last_error():
        get_last_error_msg = dll.ecos_last_error_msg
        get_last_error_msg.restype = c_char_p
        err = get_last_error_msg()
        return err.decode()

    @staticmethod
    def set_log_level(lvl: str):
        set_log_level = dll.ecos_set_log_level
        set_log_level.argtypes = [c_char_p]
        set_log_level(lvl.encode())



from typing import Callable

try:
    import matplotlib.pyplot as plt

    MATPLOTLIB_AVAILABLE = True

    import re
    import pandas as pd


except ImportError:
    MATPLOTLIB_AVAILABLE = False


class TimeSeriesConfig:

    def __init__(self, title: str, y_label: str, identifiers: list[str],
                 modifiers: dict[str, Callable[[any], any]] = None):
        self.title = title
        self.y_label = y_label
        self.identifiers = identifiers
        self.modifiers = modifiers


class XYSeriesConfig:

    def __init__(self, title: str, x_label: str, y_label: str, series: dict[str, tuple[str, str]]):
        self.title = title
        self.x_label = x_label
        self.y_label = y_label
        self.series = series


if MATPLOTLIB_AVAILABLE:
    # Plotter class for plotting time series and XY series from CSV files
    class Plotter:
        fig_id = 0

        def __init__(self, csv_file: str, configs):
            self.csv = pd.read_csv(csv_file, delimiter=r",\s+", engine="python")

            if not isinstance(configs, (list, tuple)):
                configs = [configs]

            for config in configs:
                if isinstance(config, TimeSeriesConfig):
                    self.make_time_series(self.csv, config)
                elif isinstance(config, XYSeriesConfig):
                    self.make_xy_series(self.csv, config)
                else:
                    raise Exception("Missing or unsupported configuration")

        @staticmethod
        def show():
            plt.show()

        @staticmethod
        def make_time_series(csv, timeseries: TimeSeriesConfig):

            t = csv['time']
            plt.figure("figure_{}".format(Plotter.fig_id))
            Plotter.fig_id += 1
            plt.title(timeseries.title)
            plt.xlabel("time[s]")
            plt.ylabel(timeseries.y_label)

            # Create a mapping from cleaned column names to actual column names
            clean_col_map = {}
            for col in csv.columns:
                clean_col = re.sub(r'\s*\[.*?]\s*$', '', col)  # Remove trailing type markers like [REAL], [INT]
                clean_col_map[clean_col] = col

            for identifier in timeseries.identifiers:
                if identifier in clean_col_map:
                    actual_col = clean_col_map[identifier]
                    data = csv[actual_col]
                    if timeseries.modifiers is not None and identifier in timeseries.modifiers:
                        data = data.map(timeseries.modifiers[identifier])
                    plt.plot(t, data, label=actual_col)
                else:
                    print(f"Warning: Identifier '{identifier}' not found in CSV columns.")

            plt.legend(loc='upper right')

        @staticmethod
        def make_xy_series(csv, xyseries: XYSeriesConfig):

            plt.figure("figure_{}".format(Plotter.fig_id))
            Plotter.fig_id += 1
            plt.title(xyseries.title)
            plt.xlabel(xyseries.x_label)
            plt.ylabel(xyseries.y_label)
            for name in xyseries.series:
                xy = xyseries.series[name]
                x = xy[0]
                y = xy[1]
                marker = None if len(xy) < 3 else xy[2]

                m1 = csv.columns.str.contains(re.escape(x))
                data1 = csv.loc[:, m1]
                m2 = csv.columns.str.contains(re.escape(y))
                data2 = csv.loc[:, m2]
                if marker is None:
                    plt.plot(data1, data2, label=name)
                else:
                    plt.plot(data1, data2, marker, label=name)

            plt.legend(loc='upper right')

else:
    class Plotter:
        def __init__(self, csv_file, configs):
            print("matplotlib not available. Plotting is disabled.")

        @staticmethod
        def show():
            pass

        @staticmethod
        def make_time_series(csv, timeseries: TimeSeriesConfig):
            pass

        @staticmethod
        def make_xy_series(csv, xyseries: XYSeriesConfig):
            pass




import glob
from setuptools import setup


def version():
    with open("version.txt", "r") as f:
        return f.readline().strip()


def binaries():
    return glob.glob("ecospy/binaries/ecos*") + glob.glob("ecospy/binaries/proxyfmu*")


setup(name="ecospy",
      version=version(),
      data_files=[
          ("Scripts", binaries())
      ]
      )
